{-# OPTIONS -fglasgow-exts #-}

module Plugins.FOL.Translation where

import Control.Monad.RWS
import Control.Monad.Error
import Data.Set as Set ( Set, empty, member, insert, toList, fromList )

import Syntax.Internal
import Syntax.Values

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Context
import TypeChecker.Monad.Options
import TypeChecker.Monad.MetaVars
import TypeChecker.Monad.Datatypes
import TypeChecker.Context
import TypeChecker.Equality
import TypeChecker.UnfoldConstants
import TypeChecker.Values
import TypeChecker.Values.DeBruijn
import TypeChecker.Abstractions
import TypeChecker.MetaVars
import TypeChecker.Patterns
import TypeChecker.Universes
import TypeChecker.Abstractions

import Utils.Error
import Utils.Map hiding (map, null)
import Utils.Monads
import Utils.Names		( fresh )
import Utils.Pairs
import Utils.Debug

import Plugins.FOL.Types
import Plugins.FOL.Primitive
import Plugins.FOL.Pretty
import Plugins.FOL.Constants
import Plugins.FOL.Errors
import Plugins.FOL.Computation	( unfoldAllButFOLConstants )
import Plugins.FOL.Substitution
import Plugins.FOL.Arguments

{--------------------------------------------------------------------------
    The Translation Monad
 --------------------------------------------------------------------------}

type M = RWST (Arguments, Set Name) [Formula] Int TC

runM :: Arguments -> M a -> TC (a, [Formula])
runM args m = evalRWST m (args, Set.empty) 0

isBoundVar :: Name -> M Bool
isBoundVar x = (Set.member x . snd) <$> ask

bindVar :: Name -> M a -> M a
bindVar x = local $ \(a,s) -> (a, Set.insert x s)

setNames :: [Name] -> M a -> M a
setNames xs = local $ \(a,_) -> (a, Set.fromList xs)

getNames :: M [Name]
getNames = Set.toList . snd <$> ask

getFrozen :: M [Name]
getFrozen = freeze . fst <$> ask

{--------------------------------------------------------------------------
    Lambda Lifting
 --------------------------------------------------------------------------}

lambdaName :: M String
lambdaName =
    do	n <- (get :: M Int)
	put (n + 1)
	return $ "lam_" ++ show n

lambdaLift :: Abs Value -> M FOLTerm
lambdaLift f =
    do	name	<- lambdaName
	xs	<- getNames
	x	<- lift $ fresh (absName f)
	let call = FOLFun name [] `apps` map folVar xs
	    lhs  = app [call, folVar x]
	    args = xs ++ [x]
	    args'= map nameString args
	frz	<- getFrozen
	rhs	<- bindVar x $ valueToTerm =<<
		    lift (unfoldAllButFOLConstants frz
			    $ absApp f (vId x))
	tell [forAll (length args) $ \ts -> substF args' ts $ equal [lhs, rhs]]
	return call
    where
	apps = foldl (\x y -> app [x,y])

{--------------------------------------------------------------------------
    Translation
 --------------------------------------------------------------------------}

-- | Infer the type of an argument to the plugin
inferArgument :: Term -> TC (Value, Type)
inferArgument t =
    allowUnsolvedMetas $ inferType t

{- |
    The top level function. Returns both the value of its argument and the
    corresponding first order formula.
    <p>
    The argument can be one of four things:
	* a primitive constant (e.g. __fol_extensionality)
	* proof of a proposition
	* the name of a datatype
	* the name of a function defined by pattern matching
-}
argumentToFOL :: Value -> Type -> M Formula
argumentToFOL v t =
    case vView v of
	VId x | Just p <- Prelude.lookup (nameString x) fol_primitives
		-> return p
	_	-> prop v t
    where
	prop v t
	    | isProp t	=
		do  frz <- getFrozen
		    valueToFOL =<< lift (unfoldAllButFOLConstants frz $ unEl t)
	    | otherwise	= dataType v t

	dataType v t =
	    do	(_, _, _, cs) <- lift $ getDatatype v
                -- get list of arities @as@ for constructors @cs@
		as <- lift $ map arity <$> mapM lookupDef cs
		return $ dataToFOL (zip cs as)
	    `catchError` \_ -> patternDef v t
	    where
		arity (DefVal h ar _ _ _ _) = ar - h

	-- Note: I don't understand this definition. Shouldn't we check the
	-- type before claiming that it's not a prop? No. The type is checked
	-- in the prop function above.
	patternDef v t =
	    case vAppView v of
		VAVDef x h vs
		    | length vs == h	->
			do  DefVal _ _ _ _ def t' <- lift $ lookupDef x
			    let cs = getClausesTC def
			    when (null cs) $ lift $ failure $ notOfTypePropError t (getSort t)
			    patternDefToFOL (vDef x h) t' cs
		_ -> lift $ failure $ notOfTypePropError t (getSort t)

equalityToFOL :: [Name] -> Value -> Value -> M Formula
equalityToFOL xs lhs rhs =
    setNames xs $
    do	lhsf <- valueToTerm lhs
	frz  <- getFrozen
	rhsf <- valueToTerm =<< lift (unfoldAllButFOLConstants frz rhs)
	return	$ forAll' xs
		$ \ts -> substF xs' ts $ equal [lhsf,rhsf]
    where
	xs' = map nameString xs

equivalenceToFOL :: [Name] -> Value -> Value -> M Formula
equivalenceToFOL xs lhs rhs =
    setNames xs $
    do	lhsf <- propToFOL lhs
	frz  <- getFrozen
	rhsf <- propToFOL =<< lift (unfoldAllButFOLConstants frz rhs)
	return	$ forAll' xs
		$ \ts -> substF xs' ts $ lhsf `equiv` rhsf
    where
	xs' = map nameString xs

substF xs ts f = foldr (uncurry substFormula) f $ zip xs ts

patternDefToFOL :: Value -> Type -> [ClauseValue] -> M Formula
patternDefToFOL d t cs = bigAnd . concat <$> mapM clauseToFOL cs
    where
	clauseToFOL (ClauseVal ps Nothing) = return []
	clauseToFOL (ClauseVal ps (Just v)) =
	    (:[]) <$>
	    do	vs  <- lift $ mapM evalPat ps
		let n = sum $ map nPatternVars ps
		xs  <- replicateM n $ fresh "x"
		let lhs = d `vApps` vs
		    rhs = instantiate v $ map vId $ reverse xs

		t'  <- lift $ t `piApps` vs
		if isProp t'
		    then equivalenceToFOL xs lhs rhs
		    else equalityToFOL xs lhs rhs

	evalPat VAbsurdP     = return $ vId dontCare -- TODO: hmm
	evalPat VWildP	     = return $ vId dontCare
	evalPat VIdP	     = vId <$> fresh "x"
	evalPat (VPairP p q) = vPair <$> evalPat p <*> evalPat q
	evalPat (VConP c ps) =
	    do	ps'		    <- mapM evalPat ps
		DefVal h ar _ _ _ _ <- lookupDef c
		return $ vCon c (error "FOL.Translation: won't happen") h ar `vApps` ps'

conApp :: Name -> [FOLTerm] -> FOLTerm
conApp c xs = foldl (\a b -> app [a,b]) (folFun c []) xs

forAll :: Int -> ([FOLTerm] -> Formula) -> Formula
forAll 0 f = f []
forAll n f =	ForAll "x"	$ \x ->
		forAll (n - 1)	$ \xs -> f (x:xs)

forAll' :: [Name] -> ([FOLTerm] -> Formula) -> Formula
forAll' [] f	 = f []
forAll' (x:xs) f = ForAll (nameString x) $ \x ->
		   forAll' xs $ \xs ->
		   f (x:xs)

dataToFOL :: [(Name, Arity)] -> Formula
dataToFOL cs = bigAnd $ map consToFOL (square cs)
		    ++	concatMap injective cs
    where
	consToFOL ((c,n), (c',m)) =
	    forAll n $ \xs ->
	    forAll m $ \ys ->
		Not (equal [conApp c xs, conApp c' ys])

	injective (c,0) = []
	injective (c,n) = [
		forAll n $ \xs ->
		forAll n $ \ys ->
		    equal [conApp c xs, conApp c ys] `implies`
			bigAnd (zipWith eq xs ys)
	    ]

	eq x y = equal [x,y]

	square (x:xs) = map ((,) x) xs ++ square xs
	square [] = []

valueToFOL :: Value -> M Formula
valueToFOL v =
    do  defs <- lift getDefinitions
	frozen <- getFrozen
	propToFOL =<< lift (unfoldAllButFOLConstants frozen v)

-- | Should perhaps check if the value is a subtype of Prop. But it doesn't.
isProp :: Type -> Bool
isProp t =
    case getSort t of
	VSProp  -> True
	_	-> False

isSet :: Type -> TC ()
isSet t = typeType t =<= el vsType (vSort vsSet)

piToFOL :: Type -> Abs Type -> (Value -> M Formula) -> M Formula
piToFOL t cl toFOL
--     | not (occursIn 0 b) && isProp t =
-- 	    implies <$> toFOL t <*> toFOL (raiseBy (-1) b)
--     | name == "_"   = fail "pi is neither implies not forall"
    | otherwise	    =
	do  lift $ sayWhat (higherOrderQuantification (dummyName name) t) $ isSet t
	    x <- lift $ fresh name
	    fm <- bindVar x $ toFOL (unEl $ instantiate b [vId x])
	    return $ ForAll (nameString x) $ \t -> substFormula (nameString x) t fm
    where
	name	= absName cl
	b	= absBody cl
{- the alternative @return fm@ would not work since the translation below
   does not handle implicitely quantified variables @FOLVar@ under negation
   correctly. -}

propToFOL :: Value -> M Formula
propToFOL v =
    case vAppView v of
	VAVDef c _ []
	    | nameString c == fol_True	-> return TRUE
	    | nameString c == fol_False	-> return FALSE
	VAVDef c _ [v1, v2]
	    | nameString c == fol_And	-> bin And v1 v2
	    | nameString c == fol_Or	-> bin Or v1 v2
	    | nameString c == fol_Implies	-> bin implies v1 v2
	    | nameString c == fol_Equiv	-> bin equiv v1 v2
	VAVDef c _ [t, p]
	    | nameString c == fol_ForAll	->
		do  x <- lift $ fresh =<< guessName p
		    fm <- bindVar x $ propToFOL $ vApply p (vId x)
		    return $ ForAll (nameString x) $ \t -> substFormula (nameString x) t fm
	    | nameString c == fol_Exists	->
		do  x <- lift $ fresh =<< guessName p
		    fm <- bindVar x $ propToFOL $ vApply p (vId x)
		    return $ Exists (nameString x) $ \t -> substFormula (nameString x) t fm
	    where
		guessName v =
		    do	v' <- unfoldConstants v
			case vView v' of
			    VLam cl -> return $ noWild $ absName cl
			    _	    -> return "x"
		    where
			noWild "_"  = "x"
			noWild x    = x
	VAVDef c _ [t, v1, v2]
	    | nameString c == fol_Equals   ->
		do  t' <- lift $ unfoldConstants t
		    case vView t' of
			-- Make sure all functions we compare are fully applied
			VPi w _   ->
			    do  let (v1',v2') = raise (v1,v2)
				    v1'' = vLam $ mkAbs "x" $ vApply v1' (vBId 0)
				    v2'' = vLam $ mkAbs "x" $ vApply v2' (vBId 0)
				equal <$> mapM valueToTerm [v1'',v2'']
			_   -> equal <$> mapM valueToTerm [v1,v2]

	VAVDef c _ [v1]
	    | nameString c == fol_Not	-> Not <$> propToFOL v1
	VAVDef c h vs ->
	    Predicate (nameString c) <$> mapM valueToTerm (drop h vs)    -- Ignore hidden arguments
	VAVId x vs		->
	    do  b <- isBoundVar x
		if b
		    then lift $ failure $ notAFOLPropositionError v
		    else Predicate (nameString x) <$> mapM valueToTerm vs

	VAVProj w p vs  ->	-- TODO: this translation is a bit fishy
	    case projs (VProj w p) of
		(VId x, ps)   ->
		    ifM (isBoundVar x)
			(lift $ failure $ notAFOLPropositionError v)
			(do ts <- mapM valueToTerm vs
			    return $ Predicate (nameString x ++ concatMap proj ps) ts
			)
		(VDef x h, ps) ->
		    do  ts <- mapM valueToTerm $ drop h vs
			return $ Predicate (nameString x ++ concatMap proj ps) ts
		_		-> lift $ failure $ notAFOLPropositionError v
	    where
		proj VFst	    = "_1"
		proj VSnd	    = "_2"
		proj (VField _ x)   = x	    -- TODO: handle hiding

		projs (VProj v p) = (w, ps ++ [p])
		    where
			(w,ps) = projs (vView v)
		projs v	= (v, [])

	VAVSigma w f    ->
	    do	x <- lift $ fresh $ absName f
		bin And (unEl w) $ unEl $ absApp f $ vId x
	VAVPi v cl	    -> piToFOL v cl propToFOL
	_		    -> lift $ failure $ notAFOLPropositionError v
    where
	bin op v1 v2 = op <$> propToFOL v1 <*> propToFOL v2

-- TODO: use the uid
folVar x = FOLVar (nameString x)
folFun x = FOLFun (nameString x)

valueToTerm :: Value -> M FOLTerm
valueToTerm v =
    case vAppView v of
	VAVId x vs	->
	    do  bound <- isBoundVar x
		ts <- mapM valueToTerm vs
		return $ foldl bapp (if bound then folVar x else folFun x []) ts
			
	VAVDef x h vs	->
	    do	ts <- mapM valueToTerm $ drop h vs
		return $ foldl bapp (folFun x []) ts
	VAVCon x _ h _ vs ->
	    do	ts <- mapM valueToTerm $ drop h vs
		return $ foldl bapp (folFun x []) ts
	VAVPair v1 v2	-> pair <$> mapM valueToTerm [v1,v2]
	VAVProj v p vs	->
	    do	ts <- mapM valueToTerm vs
		t <- valueToTerm v
		return $ foldl bapp (proj [t, projToTerm p]) ts
	VAVSigma w f    ->
	    do	x <- lift $ fresh (absName f)
		let b = absApp f $ vId x
		sigma <$> mapM valueToTerm [unEl w, unEl b]
	VAVPi w f	->
	    do	x <- lift $ fresh (absName f)
		let b = absApp f $ vId x
		fun <$> mapM valueToTerm [unEl w, unEl b]
	v@(VAVLam f)	-> lambdaLift f
	VAVTop		-> return $ FOLFun "Top_" []
	_		-> lift $ failure $ notAFOLTermError v
    where
	bapp x y = app [x,y]

projToTerm :: ProjectionValue -> FOLTerm
projToTerm VFst		= FOLFun kwFst []
projToTerm VSnd		= FOLFun kwSnd []
projToTerm (VField _ x) = FOLFun x []

