{-# OPTIONS -cpp #-}
{-# OPTIONS -fglasgow-exts #-}

module Syntax.Internal
	( module Syntax.Name
	, module Syntax.Internal
	)
    where

import Data.Generics
import Data.Maybe
import Data.List
import Data.IORef
import Debug.Trace

import Utils.Map    ( Map )
import Utils.Pairs
import Utils.Generics
import Utils.Lists
import Utils.Ref
import Utils.Monads

import Syntax.Name
import Syntax.Position
import Syntax.Scope

-- * The internal syntax 
----------------------------------------------------------------------

-- ** Terms

data Term
	= App Term Term
	| HApp Term Term        -- ^ hidden application
	| Var Range Name	-- ^ name bound by lambda or let
	| Def Range Name	-- ^ defined names
	| Con Range Name	-- ^ data constructors
	| Data Range Name	-- ^ datatypes
	| Sig Range Name	-- ^ signatures
	| Lam Range Pattern Term
		-- ^ Only patterns of the form @x | (p,q)@ are allowed.
	| Pair Range Term Term
	| Project Range Term Projection
		    -- ^ invariant: terms annotated with types shouldn\'t contain
		    -- any 'Field' projections (they should be translated to
		    -- 'Fst' and 'Snd's).
	| Hash Range Universe
	| ISet Range -- ^ impredicative set
	| Set Range
	| Prop Range
	| Type Range
	| Pi Range Pattern Term	Term
		-- ^ Only patterns of the form @x | (p,q)@ are allowed.
	| Sigma Range Name Term Term
	| Struct Range [(Name, [Pattern], [Pattern], Term)]
	| CallPlugin Range ScopeEnv String [Term]
		-- ^ We don't require the arguments to a plugin to
		--   be type checked.
	| Let Range [Definition] Term
	| RewriteIn Range [Term] Term
	| Meta Range Name
	| Absurd Range	-- ^ Only used in patterns
    deriving (Typeable, Data)

-- ** Universes

data Universe = UZero | USucc Universe
    deriving (Typeable, Data, Eq)

intToUniverse :: Int -> Universe
intToUniverse 0 = UZero
intToUniverse n = USucc $ intToUniverse (n - 1)

uValue :: Universe -> Int
uValue UZero	    = 0
uValue (USucc u)    = 1 + uValue u

-- ** Projections

-- | The ranges are used in texification. Only strictly necessary for the 'Field'
--   constructor.
data Projection	= Proj Range Int | Field Range String
    deriving (Typeable, Data)

type Telescope = [(Name, Term)]

{- | @telePi (xs::As) t = (xs::As) -> t@ -}
telePi :: Telescope -> Term -> Term
telePi tel t = foldr (\(x,a) b -> Pi (fuseRange x b) (VarP x) a b) t tel

{- | @teleLam (xs::As) t = \ xs -> t@ -}
teleLam :: Telescope -> Term -> Term
teleLam tel = buildLam $ map (var . fst) tel
    where
	var x = VarP x

{- | @buildLam [x1,...,xn] t = \ x1 -> ... \ xn -> t@ -}
buildLam :: [Pattern] -> Term -> Term
buildLam ps e = foldr (\p b -> Lam (fuseRange p b) p b) e ps

-- ** Definitions

type Program = [Definition]

-- | We've decided to merge constants and definitions.
-- We will only unfold in typechecking and equality
-- testing.
-- The syntax for the new definitions is
--   foo :: (x0::A0) -> .. -> (xn::An) -> T
--   foo x0 .. xn = e
-- That is, we drop telescopes altogether. It also
-- means that definitions have arities and that unfolding
-- is only done on fully applied things.
--
-- You can leave out the type signature if the body is
-- inferrable.
type IsSized = Bool
data Definition
	= NormalDef Range Name Hidden Arity Term [Clause]
  -- ^ the name of each clause coincides with the name of the definition 
        | NewAxiom Range Name Hidden Term
	| Synonym Range Name Term
	| NewData IsSized Range Name Telescope Telescope Term [Constructor]
	| NewSig Range Name Telescope Telescope Term [Name] [Constructor]
	| Rewrite Range [Term]
	| Mutual Range [Definition]
	| AbstractDefs Range [Definition]
    deriving (Typeable, Data)

type Hidden = Int
type Arity  = Int

-- ** Data type and constructor definitions

{- | Constructor declarations @c :: C@.  The type @C@ of the
constructor is inside the scope of the telescope @(xs::As)@ of the
data declaration.  @C@ must end in @D xs@. (Otherwise we would have
inductive families.) -}
data Constructor = Constructor Range Name Telescope Term
    deriving (Typeable, Data)

constructorName (Constructor _ x _ _) = x

data Clause = Clause Range Name [Term] WithClause
    deriving (Typeable, Data)

data WithClause	= NoWith Term
		| With Name [Term] [Clause]
		| AbsurdClause
    deriving (Typeable, Data)

-- | Patterns @p@ are terms composed of constructors @c@ and variables @x@.
data Pattern
	= VarP Name		-- ^ . @x@
	| ConP Range Name [Pattern] [Pattern]
				-- ^ . @c ps@
	| PairP Range	Pattern
			Pattern
				-- ^ . @(p1,p2)@
	| WildP Range		-- ^ . @_@
	| AbsurdP Range		-- ^ . @()@
    deriving (Typeable, Data)

-- * Views on terms 
----------------------------------------------------------------------

data TAppView
	= TAVar  Range Name        [Term]   -- ^ variables can't have hidden arguments
	| TACon  Range Name [Term] [Term]
	| TADef  Range Name [Term] [Term]
	| TAData Range Name [Term] [Term]
	| TASig  Range Name [Term] [Term]
	| TAProj Range Term String [Term] [Term]    -- ^ named projection
	| TAOther Term

tAppView :: Term -> TAppView
tAppView t =
    case t of
	App t t'		-> tAppView t `app` t'
	HApp t t'		-> tAppView t `happ` t'
	Var r x			-> TAVar r x []
	Con r x			-> TACon r x [] []
	Def r x			-> TADef r x [] []
	Data r x		-> TAData r x [] []
	Sig r x			-> TASig r x [] []
	Project r e (Field _ x) -> TAProj r e x [] []
	_			-> TAOther t
    where
	app (TAVar  r x	     ts) t = TAVar  r x (ts ++ [t])
	app (TACon  r x   hs ts) t = TACon  r x hs (ts ++ [t])
	app (TADef  r x   hs ts) t = TADef  r x hs (ts ++ [t])
	app (TAData r x   hs ts) t = TAData r x hs (ts ++ [t])
	app (TASig  r x   hs ts) t = TASig  r x hs (ts ++ [t])
	app (TAProj r e x hs ts) t = TAProj r e x hs (ts ++ [t])
	app (TAOther t')	 t = TAOther (App t' t)

	happ (TACon  r x   hs []) t = TACon r x (hs ++ [t]) []
	happ (TADef  r x   hs []) t = TADef r x (hs ++ [t]) []
	happ (TASig  r x   hs []) t = TASig r x (hs ++ [t]) []
	happ (TAData r x   hs []) t = TAData r x (hs ++ [t]) []
	happ (TAProj r e x hs []) t = TAProj r e x (hs ++ [t]) []
	happ ta			  t = TAOther $ HApp (unTAppView ta) t

unTAppView :: TAppView -> Term
unTAppView (TAVar  r x    ts) = foldl App (Var r x) ts
unTAppView (TACon  r x hs ts) = foldl App (foldl HApp (Con  r x) hs) ts
unTAppView (TADef  r x hs ts) = foldl App (foldl HApp (Def  r x) hs) ts
unTAppView (TASig  r x hs ts) = foldl App (foldl HApp (Sig  r x) hs) ts
unTAppView (TAData r x hs ts) = foldl App (foldl HApp (Data r x) hs) ts
unTAppView (TAProj r e x hs ts) = foldl App (foldl HApp (Project r e (Field r x)) hs) ts
unTAppView (TAOther t)	      = t

-- Syntactic equality on terms --------------------------------------

-- | This is the literal equality, ignoring position information
--   and annotations.
instance Eq Term where
    App a b		== App c d		= (a,b)		== (c,d)
    Var _ x		== Var _ y		= x		== y
    Def _ x		== Def _ y		= x		== y
    Con _ x		== Con _ y		= x		== y
    Data _ x		== Data _ y		= x		== y
    Lam _ x a		== Lam _ y b		= (x,a)		== (y,b)
    Hash _ u		== Hash _ v		= u		== v
    Set _		== Set _		= True
    Prop _		== Prop _		= True
    Type _		== Type _		= True
    Pi _ p a b		== Pi _ q c d		= (p,a,b)	== (q,c,d)
    CallPlugin _ _ x ts	== CallPlugin _ _ y us	= (x,ts)	== (y,us)
    Let _ defs a	== Let _ defs' b	= (defs,a)	== (defs',b)
    RewriteIn _ a b	== RewriteIn _ c d	= (a,b)		== (c,d)
    Pair _ a b		== Pair _ c d		= (a,b)		== (c,d)
    Project _ a p	== Project _ b q	= (a,p)		== (b,q)
    Sigma _ x a b	== Sigma _ y c d	= (x,a,b)	== (y,c,d)
    Meta _ x		== Meta _ y		= x		== y
    _			== _			= False

instance Eq Definition where
    NormalDef _ x h a e cs	== NormalDef _ y h' a' e' cs'	    = (x,h,a,e,cs)	== (y,h',a',e',cs')
    NewAxiom _ x h e		== NewAxiom _ y h' e'		    = (x,h,e)		== (y,h',e')
    Synonym _ x a		== Synonym _ y b		    = (x,a)		== (y,b)
    NewData _ _ x htel tel a cs == NewData _ _ y htel' tel' b cs'   = (x,htel,tel,a,cs) == (y,htel',tel',b,cs')
    NewSig _ x htel tel a xs cs == NewSig _ y htel' tel' b ys cs'   = (x,htel,tel,a,xs,cs) == (y,htel',tel',b,ys,cs')
    Rewrite _ as		== Rewrite _ bs			    = as		== bs
    Mutual _ as			== Mutual _ bs			    = as		== bs
    AbstractDefs _ as		== AbstractDefs _ bs		    = as		== bs
    _				== _				    = False

instance Eq Clause where
    Clause _ x as with == Clause _ y bs with'   = (x,as,with) == (y,bs,with')

instance Eq WithClause where
    NoWith a	== NoWith b	= a	   == b
    With x as c	== With y bs d	= (x,as,c) == (y,bs,d)
    _		== _		= False

instance Eq Constructor where
    Constructor _ x tel a == Constructor _ y tel' b = (x,tel,a) == (y,tel',b)

instance Eq Pattern where
    WildP _	    == WildP _		= True
    VarP x	    == VarP y		= x == y
    ConP _ c hs ps  == ConP _ d is qs	= (c,ps,hs) == (d,qs,is)
    PairP _ p q	    == PairP _ r s	= (p,q) == (r,s)
    _		    == _		= False

instance Eq Projection where
    Proj _ n	== Proj _ m	= n == m
    Field _ x	== Field _ y	= x == y
    _		== _		= False

-- Ranges ------------------------------------------------------------

instance HasRange Term where
    getRange (App t u)		    = fuseRange t u
    getRange (HApp t u)		    = fuseRange t u
    getRange (Var r _)		    = r
    getRange (Def r _)		    = r
    getRange (Con r _)		    = r
    getRange (Sig r _)		    = r
    getRange (Data r _)		    = r
    getRange (Lam r _ _)	    = r
    getRange (Hash r _)		    = r
    getRange (Set r)		    = r
    getRange (ISet r)		    = r
    getRange (Prop r)		    = r
    getRange (Type r)		    = r
    getRange (Pi r _ _ _)	    = r
    getRange (CallPlugin r _ _ _)   = r
    getRange (Let r _ _)	    = r
    getRange (RewriteIn r _ _)	    = r
    getRange (Pair r _ _)	    = r
    getRange (Sigma r _ _ _)        = r
    getRange (Project r _ _)	    = r
    getRange (Meta r _)		    = r
    getRange (Struct r _)	    = r
    getRange (Absurd r)		    = r

instance HasRange Projection where
    getRange (Proj r _)	    = r
    getRange (Field r _)    = r

instance HasRange Pattern where
    getRange (VarP x)	    = getRange x
    getRange (ConP r _ _ _) = r
    getRange (PairP r _ _)  = r
    getRange (WildP r)	    = r
    getRange (AbsurdP r)    = r

